#include <cstddef>
using std::size_t;

#include <cassert>

#include <cstdlib>
using std::exit;

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <iomanip>
using std::setw;
using std::setfill;

#include <fstream>
using std::ifstream;
using std::ofstream;

#include <string>
using std::string;

#include <sstream>
using std::stringstream;

#include "grid.hpp"

ifstream fin;
ofstream fout1;
ofstream fout2;

enum methodFlag { JACOBI, GAUSS, CHECKERBOARD, WEIGHTED };

void bad(void)
{
	cerr << "Usage: relax jacobi|gauss|checkboard|(weighted #) phi|E filename" << endl; 
	exit(1);	
}

int main(int argc, char* argv[])
{
	if (argc < 2) bad();
	const string option1 = argv[1];
	if (!(option1 != "weighted" and argc == 4 or option1 == "weighted" and argc == 5)) bad();
	

	methodFlag method;
	size_t shift = 0;
	double a;
	if (option1 == "jacobi")
		method = JACOBI;
	else if (option1 == "gauss")
		method = GAUSS;
	else if (option1 == "checkerboard")
		method = CHECKERBOARD;
	else if (option1 == "weighted")
	{
		method = WEIGHTED;
		shift = 1;
		a = atof(argv[2]);
		if (a < 0.0)
		{
			cerr << "Weighting value a=" << a << " must be positive." << endl;
			exit(1);
		}
	}
	else
	{
		cerr << "Did not recognize option `" << option1 << "`" << endl;
		exit(1);
	}
	
	const string option2 = argv[2+shift];
	const string filename = argv[3+shift];
	fin.open(filename.c_str());
	if (!fin.is_open())
	{
		cerr << "Could not open file `" << filename << "`." << endl;
		exit(1);
	}
	
	
	size_t Lx, Ly;
	fin >> Lx >> Ly;
	assert(fin.good());
	grid G(Lx,Ly);

	if (option2 == "phi")
		G.output_gradient = false;
	else if (option2 == "E")
		G.output_gradient = true;
	else
	{
		cerr << "Did not recognize option `" << option2 << "`" << endl;
		exit(1);
	}
	
	// # dead cell
	// . free cell
	// 0 fixed to value zero
	// X fixed to specified value
	// + positive charge
	// - negative charge
	vector<size_t> Xcolumn;
	for (size_t y = 0; y < Ly; ++y)
	{
		Xcolumn.resize(0);
		for (size_t x = 0; x < Lx; ++x)
		{
			char c;
			fin >> c;
			assert(fin.good());
			if (c == '.') ; // do nothing
			else if (c == '#')
				G.mark_as_dead(x,y);
			else if (c == '0')
			{
				G(x,y) = 0.0;
				G.fix(x,y);
			}
			else if (c == 'X')
			{
				Xcolumn.push_back(x);
				G.fix(x,y);
			}
			else if (c == '+')
				G.charge(x,y) = 1.0;
			else if (c == '-')
				G.charge(x,y) = -1.0;
			else
				assert(false);
		}
		for (size_t n = 0; n < Xcolumn.size(); ++n)
			fin >> G(Xcolumn[n],y);
		assert(y == Ly-1 or fin.good());
	}

	fout1.open("output0000.dat");
	fout1 << G << endl;
	fout1.close();

	const string range0 = ( G.output_gradient ? "set zrange [] writeback" : "set yrange [] writeback" );
	const string range = ( G.output_gradient ? "set zrange restore" : "set yrange restore" );

	stringstream s;
	if (!G.output_gradient)
		s << "s";
	s << "plot[" << 0 << ":" << Lx-1 << "][" << 0 << ":" << Ly-1 << "] '";
	const string plot_begin = s.str();
	const string plot_end = ( G.output_gradient ? "' using 1:2:($3/dx):($4/dx) with vector" : "' matrix" );
	
	fout2.open("movie.gp");
	if (!G.output_gradient)
		fout2 << "set pm3d map" << endl;
	else
		fout2 << "dx = 0.1" << endl;
	fout2 << "set size ratio 1" << endl
	      << range0 << endl
	      << plot_begin << "output0000.dat" << plot_end << endl
	      << "pause -1 \"Press return to start\"" << endl;

	/* from gnuplot help:
	** The `vectors` style draws a vector from (x,y) to (x+xdelta,y+ydelta).  Thus
	** it requires four columns of data.  It also draws a small arrowhead at the
	** end of the vector.
	**
	** Example:
	**  plot 'file.dat' using 1:2:3:4 with vectors head filled lt 2
	**
	** `set clip one` and `set clip two` affect drawing vectors.
	** Please see `set clip` and `arrowstyle`.
	*/

	unsigned int passes = 0;
	do
	{
		G.zero_error();
		if (method == JACOBI)
		{
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					G.relax_tentative(x,y);
			G.relax_commit();
		}
		else if (method == GAUSS)
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					G.relax(x,y);
		else if (method == CHECKERBOARD)
		{
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					if (passes%2 == (x+y)%2)
						G.relax(x,y);
		}
		else if (method == WEIGHTED)
		{
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					G.relax_tentative(x,y);
			G.relax_commit(a);
		}
		else
			assert(false);
		
		++passes;

		stringstream s;
		s << "output" << setw(4) << setfill('0') <<  passes << ".dat";
		fout1.open(s.str().c_str());
		fout1 << G << endl;
		fout1.close();
		
		fout2 << range << endl << plot_begin << s.str() << plot_end << endl << "pause 0.05" << endl;

		cout << "Error : " << G.error() << endl;
	} while (G.error() > 1E-5 and passes < 10000);

	if (passes == 10000)
		cout << "Did not converge after 10000 passes" << endl;
	else
		cout << "Converged in " << passes << " passes" << endl;

	fout2.close();	
	return 0;
} 
